#!/usr/bin/env php
<?php
  date_default_timezone_set ("Europe/Rome");
  set_error_handler("customError");
  const CACHEDIR = "./cache";

  if (!file_exists(CACHEDIR)){
      mkdir(CACHEDIR);
  }
  
  function xpath1 ($xml, $xpath){
    $nodes = $xml->xpath($xpath);
    if ($nodes && count($nodes)){
      return (string) $nodes[0];
    }
    return null;
  }

  function customError($errno, $errstr)  {
    echo "Error: errno [$errno]; errstr [$errstr]";
    //die;
  }

  function curlIt ($url,$dest){
    $curl = curl_init();

    // set URL and other appropriate options
    curl_setopt($curl, CURLOPT_URL, $url);
    curl_setopt($curl, CURLOPT_FAILONERROR, true);
    curl_setopt($curl, CURLOPT_RETURNTRANSFER, true);
    curl_setopt($curl, CURLOPT_HEADER, 0);
    //curl_setopt($curl, CURLOPT_USERAGENT, '');

    // grab URL
    alog('Getting ' . $url);

    $response = curl_exec($curl);
    $httpCode = curl_getinfo($curl, CURLINFO_HTTP_CODE);
    if ($httpCode!=200){
      alog ("server report $httpCode for url $url");
    }

    // close CURL resource, and free up system resources
    curl_close($curl);

    file_put_contents ($dest.".tmp",$response);
    rename($dest.".tmp",$dest);
    
    return $response;

  }

  function httpFetch ($url,$expire=86400) {
    $cachedir = CACHEDIR."/";
    $md5cachefile = $cachedir.md5 ($url);

    alog ("md5cachefile is $md5cachefile");
    
    // check cache
    if (file_exists($md5cachefile ) && (time() - filemtime($md5cachefile )) < $expire) {
      // simply return the file
      alog ("serving cached data from file $md5cachefile for url $url");
      return file_get_contents($md5cachefile);
    }
    return curlIt ($url,$md5cachefile);
    

  }


  /**
   * parse exif time
   * @param   string   datetime as found in exif tag
   * @return time datetime
   */
  function parseExifDateTime($datetime){
    if (preg_match("/([0-9]{4}):([0-9]{2}):([0-9]{2}) ([0-9]{2}):([0-9]{2}):([0-9]{2})/",$datetime,$matches)){
      return mktime ($matches[4],$matches[5],$matches[6],$matches[2],$matches[3],$matches[1]);
    } else return null;
  }

  /**
   * parse a fraction ... e.g "1/2"
   * @param   string   fraction
   * @return double result of fraction
   */
  function parseFraction($fraction){
    if (preg_match("/([0-9.]+)\/([0-9.]+)/",$fraction,$matches)){
      $num=(double)$matches[1];
      $den=(double)$matches[2];
      return $num/$den;
    } else return null;
  }

  function ConvertToFraction($v, &$n, &$d){
    $MaxTerms = 15;         //Limit to prevent infinite loop
    $MinDivisor = 0.000001; //Limit to prevent divide by zero
    $MaxError = 0.00000001; //How close is enough

    $f = $v; //Initialize fraction being converted
    $n_un = 1; //Initialize fractions with 1/0, 0/1
    $d_un = 0;
    $n_deux = 0;
    $d_deux = 1;

    for ($i = 0;$i<$MaxTerms;$i++) {
      $a = floor($f); //Get next term
      $f = $f - $a; //Get new divisor
      $n = $n_un * $a + $n_deux; //Calculate new fraction
      $d = $d_un * $a + $d_deux;
      $n_deux = $n_un; //Save last two fractions
      $d_deux = $d_un;
      $n_un = $n;
      $d_un = $d;

      if ($f < $MinDivisor) //Quit if dividing by zero
        break;

      if (abs($v - $n / $d) < $MaxError)
        break;
      $f = 1 / $f; //Take reciprocal
    }
  }


  /**
   * get Coordinate
   * @param   array  of string degree, minutes, seconds
   * @return double coordinate
   */
  function getCoordAsDouble($coordarray){
    $deg = parseFraction ($coordarray[0]);
    $min = parseFraction ($coordarray[1]);
    $sec = parseFraction ($coordarray[2]);
    // 1 min = 1/60 of deg
    return $deg+$min/60.0+$sec/3600.0;

  }

  /**
   * return Latitude, north emisphere positive
   * @param   array of string  degree, minutes, seconds
   * @param string where N or S
   * @return double coordinate, positive values north
   */
  function getLatitudeAsDouble($coord ,$where){
    $sign = $where=='S'?-1:1;		// S or N
    return $sign* getCoordAsDouble($coord);
  }
  /**
   * return Longitude, east of Greenwich positive
   * @param   array of string  degree, minutes, seconds
   * @param string where W or E
   * @return double coordinate, positive values east of GW
   */
	function getLongitudeAsDouble($coord ,$where){
    $sign = $where=='W'?-1:1; // W or E
    return $sign* getCoordAsDouble($coord);
	}
  function alog ($message){
    echo $message."\n";
  }
 
  function doImage(){
    global $outImage;
  	global $image;
  	global $jpegQuality;
  	global $template;
  	global $dpi;
    global $rasterizer;
    global $rename;
    global $scan;
    global $outDir;
    global $geonamesAccount;
    global $dateTimePattern;
    global $photo_width;
    global $photo_height;
  	
  	alog ("input image [$image], outputimage [$outImage], dpi [$dpi] template [$template], quality [$jpegQuality] rasterizer [$rasterizer]");
		alog ("Processing [$image]");
	
	
		$exif = exif_read_data ($image);
	
		$DateTime=false;
		$DateTimeFormatted=false;

		if (array_key_exists ( "DateTime" , $exif )){
			$DateTime=parseExifDateTime($exif["DateTime"]);
			alog ("DateTime: $DateTime");			
		} 
    if (!$DateTime){
      if ($dateTimePattern) {
        // try to extract datetime from filename
        if (preg_match($dateTimePattern,$image,$matches)){
          $year = $matches["year"];
          $month = $matches["month"];
          $day = $matches["day"];
          $hour = $matches["hour"];
          $minute = $matches["minute"];
          $second = $matches["second"];
          alog ("DateTime from file name [$image] (YYYY/MM/DD HH:MM:SS) $year/$month/$day $hour:$minute:$second");
          $DateTime = mktime ($hour,$minute,$second,$month,$day,$year); 
        } 
      } 
    }
    if (!$DateTime){
		  if (array_key_exists ( "FileDateTime" , $exif )){          
        $DateTime = $exif["FileDateTime"];
      }
    }
    if (!$DateTime){
      $DateTime=time();
    }      
    $DateTimeFormatted = date ("d/m/Y H:i",$DateTime);
		alog ("DateTime: $DateTimeFormatted");

////
    // camera make
    if (array_key_exists ('Make',$exif)){
      $exifMake = $exif['Make'];
    }
    // camera model
    if (array_key_exists ('Model',$exif)){
      $exifModel = $exif['Model'];
    }
    // exposure Time
    $exposureTime=false;
    if (array_key_exists ('ExposureTime',$exif)){
      $exifExposureTime = $exif['ExposureTime'];
      $expfp = parseFraction($exifExposureTime);
      ConvertToFraction($expfp,$nu,$de);
      $exposureTime = $nu."/".$de;
      alog ("exposureTime [$exposureTime]");
    }

    // f number unformatted
    $exifFNumber=false;
    if (array_key_exists ('FNumber',$exif)){
      $exifFNumber = $exif['FNumber'];
      alog ("exifFNumber  [$exifFNumber]");
    }
    $apertureFNumber=false;
    if (array_key_exists ('COMPUTED',$exif)){
      $computed = $exif['COMPUTED'];
      if (array_key_exists ('ApertureFNumber',$computed)){
        $apertureFNumber = $computed['ApertureFNumber'];
        alog ("apertureFNumber  [$apertureFNumber]");
      }
    }

    // ISOSpeedRatings
    $ISOSpeedRatings=false;
    if (array_key_exists ('ISOSpeedRatings',$exif)){
      $exifISOSpeedRatings = $exif['ISOSpeedRatings'];
      // TODO: doesn't work in some cameras
      $ISOSpeedRatings = $exifISOSpeedRatings;
      alog ("ISOSpeedRatings  [$ISOSpeedRatings]");
    }
    // Focal length
    $focalLength=false;
    $exifFocalLength=false;
    if (array_key_exists ('FocalLength',$exif)){
      $focalLength = parseFraction($exif['FocalLength']);
      $exifFocalLength = $exif['FocalLength'];
      alog ("focalLength  [$focalLength]");
    }
    // Shutter Speed
    $shutterSpeed=false;
    if (array_key_exists ('ShutterSpeedValue',$exif)){
      $exifShutterSpeed = $exif['ShutterSpeedValue'];
      $data = parseFraction($exifShutterSpeed);      
      $data = exp($data * log(2)); 
      if ($data > 1) $data = floor($data);  //Drop the decimal.
      if ($data > 0) {
        $data = 1/$data; //Final calculation. We now have a floating number. Transform it in a pretty number
        $n=0;$d=0;
        ConvertToFraction($data, $n, $d); 
        if ($n>=1 && $d==1) 
          $data = $n." sec"; //To avoid exposure times style 3/1 sec.
        else 
          $data = $n."/".$d." sec";
      } else {
        $data = "Bulb";
      }
      $shutterSpeed = $data;
      alog ("shutterSpeed [$shutterSpeed]");
    }
    
    // Flash
    $flash=false;
    if (array_key_exists ('Flash',$exif)){
      $exifFlash = $exif['Flash'];
      $data= $exifFlash;
      // TODO: i18n
      if($data==0) $data = "No Flash";
      else if($data==1) $data = "Flash";
      else if($data==5) $data = "Flash, strobe return light not detected";
      else if($data==7) $data = "Flash, strob return light detected";
      else if($data==9) $data = "Compulsory Flash";
      else if($data==13) $data = "Compulsory Flash, Return light not detected";
      else if($data==15) $data = "Compulsory Flash, Return light detected";
      else if($data==16) $data = "No Flash";
      else if($data==24) $data = "No Flash";
      else if($data==25) $data = "Flash, Auto-Mode";
      else if($data==29) $data = "Flash, Auto-Mode, Return light not detected";
      else if($data==31) $data = "Flash, Auto-Mode, Return light detected";
      else if($data==32) $data = "No Flash";
      else if($data==65) $data = "Red Eye";
      else if($data==69) $data = "Red Eye, Return light not detected";
      else if($data==71) $data = "Red Eye, Return light detected";
      else if($data==73) $data = "Red Eye, Compulsory Flash";
      else if($data==77) $data = "Red Eye, Compulsory Flash, Return light not detected";
      else if($data==79) $data = "Red Eye, Compulsory Flash, Return light detected";
      else if($data==89) $data = "Red Eye, Auto-Mode";
      else if($data==93) $data = "Red Eye, Auto-Mode, Return light not detected";
      else if($data==95) $data = "Red Eye, Auto-Mode, Return light detected";
      else $data = "Unknown: ".$data;
      $flash = $data;
      alog ("flash [$flash]");
    }

		// check orientation
		$Orientation=1;
		if (array_key_exists ('Orientation',$exif)){
			$Orientation = $exif['Orientation'];
      alog ("Orientation [$Orientation]");
		}

//////

	
		$hasGPS=false;
	
		if (array_key_exists ('GPSLatitude',$exif) &&
				array_key_exists ('GPSLatitudeRef',$exif) &&
				array_key_exists ('GPSLongitude',$exif) &&
				array_key_exists ('GPSLongitudeRef',$exif)){
			$hasGPS=true;
			$latitude = getLatitudeAsDouble($exif['GPSLatitude'] ,$exif['GPSLatitudeRef']) ;
			$longitude = getLongitudeAsDouble($exif['GPSLongitude'] ,$exif['GPSLongitudeRef']) ;
		} else $hasGPS=false;
	
		$hasGPSDest=false;
	
		if (array_key_exists ('GPSDestLatitude',$exif) &&
				array_key_exists ('GPSDestLatitudeRef',$exif) &&
				array_key_exists ('GPSDestLongitude',$exif) &&
				array_key_exists ('GPSDestLongitudeRef',$exif)){
			$destLatitude = getLatitudeAsDouble($exif['GPSDestLatitude'] ,$exif['GPSDestLatitudeRef']) ;
			$destLongitude = getLongitudeAsDouble($exif['GPSDestLongitude'] ,$exif['GPSDestLongitudeRef']) ;
			$hasGPSDest=true;
		} else $hasGPSDest=false;
	
		$place=false;
		if ($hasGPS && $geonamesAccount){
			alog ("Latitude [$latitude], Longitude [$longitude]");
			$url = "http://api.geonames.org/findNearbyPlaceName?lat=${latitude}&lng=${longitude}&style=FULL&lang=it&username=${geonamesAccount}";
			$data = httpFetch ($url);
			$geodata = new SimpleXMLElement($data);
			$geo_name = xpath1 ($geodata,"//name");
			$geo_adminName3 = xpath1 ($geodata,"//adminName3");
			$geo_adminName2 = xpath1 ($geodata,"//adminName2");
			$geo_adminName1 = xpath1 ($geodata,"//adminName1");
			alog( "$geo_name - $geo_adminName3 - $geo_adminName2 - $geo_adminName1" );
			$place=$geo_name;$lname=$geo_name;
			if ( $geo_adminName3 && $lname!=$geo_adminName3 ) {
				$place.=", ${geo_adminName3}";$lname=$geo_adminName3;
			}
			if ( $geo_adminName2 && $lname!=$geo_adminName2 ) {
				$place.=", ${geo_adminName2}";$lname=$geo_adminName2;
			}
			if ( $lname!=$geo_adminName1 ) {
				$place.=", ${geo_adminName1}";$lname=$geo_adminName1;
			}
			alog ("Place [$place]");
		}
		$source_img_sizes = getimagesize($image);
		//print_r ($source_img_sizes);
		list($source_width, $source_height)	 = $source_img_sizes;
		alog ("source width [$source_width] source height [$source_height]" ) ;
	
		// include template
		// setup variables
		$cm=$dpi/2.54;
		$mm=$dpi/25.4;
		$in=$dpi;
		$width=$source_width;
		$height=$source_height;
    

		ob_start();
		require ("templates/${template}.svg.php");
		$buffer = ob_get_clean();
	
		file_put_contents('x.svg', $buffer);
	
	
		// convert -font monaco.ttf x.svg 10x15.jpg
		// print_r ($exif);
	
    if ($rename) {
      $oFile= date ("Y-m-d_H-i-s\.\j\p\g",$DateTime);
    } else $oFile = $outImage;
    $oFile="${outDir}/${oFile}";
    $imcnt=1;
    while ($rename && file_exists($oFile)){
      $oFile= date ("Y-m-d_H-i-s",$DateTime);
      $oFile=$oFile."_".$imcnt.".jpg";
      $oFile="${outDir}/${oFile}";      
      $imcnt++;
    }

    $command=false;
    switch ($rasterizer){
      case 'magick':
    		$command= "convert -quality ${jpegQuality} x.svg ${oFile}";
        break;
      case 'batik':
        $q = $jpegQuality/100;
        $command="rasterizer -m image/jpeg -q ${q} -d ${oFile} x.svg";
        break;
    }
    if ($command) {
      alog ("Executing [$command]");
      system ($command);
    }
  
  }


  function doDir(){
    global $inDir,$outDir;
    global $image,$outImage;
    global $recursive;

    if ($inDir==$outDir){
      echo "inDir and outDir are the same!";
      die;
    }
  	if (! file_exists($outDir)) {
  		mkdir ($outDir,0755,true);
  	}
  	// scan input dir for image files
  	$filez = scandir ( $inDir);
  	foreach ($filez as $file){
      if ($file=='.'|| $file=='..') continue;
  		$infile="${inDir}/${file}";
      if ($recursive){
        if (is_dir ($infile)){
          alog ("Descending [$infile]");
          $saveInDir = $inDir;
          $inDir = "${inDir}/${file}";
          doDir();
          $inDir = $saveInDir;
        }
      } 
  		if (preg_match("/\.jpe?g$/i",$infile)){  			
        $image=$infile;
  			$outImage=$file;
  			alog ("Processing $image");
  			doImage();
  		}
  	}
  }

  function usage (){
    global $argv;
    global $jpegQuality,$dpi,$photo_width,$photo_height,$defaultTemplate;
    echo "usage: ".$argv[0]." [OPTION]\n";
    echo "  -i, --infile=<infile>               set input file\n";
    echo "  -I, --indir=<input directory>       set input directory for scanning (default current)\n";
    echo "  -o, --outfile=<outfile>             set output file\n";
    echo "  -O, --outdir=<output directory>     set output directory (default current)\n";    
    echo "      --rename                        rename files after annotation (YYYY-MM-DD_HH-MM-SS.jpg)\n";
    echo "      --scan                          search for image files in input directory\n";
    echo "      --recursive                     do a recursive scan\n";
    echo "      --recurse                       alias for recursive\n";
    echo "  -S, --size=<photoHeight[xX]photoWidth> photo size in cm (default ${photo_height}x${photo_width} cm)\n";    
    echo "  -d, --dpi=<output dpi>              output dpi (default ${dpi})\n";    
    echo "  -r, --rasterizer=<rasterizer>       external svg to jpeg converter: magick | batik, default magick\n";         
    echo "  -q, --quality=<jpeg quality>        jpeg quality (default ${jpegQuality})\n";         
    echo "  -t, --template=<input template>     .svg.php input template name (default ${defaultTemplate})\n";
    echo "      --list-templates                list available templates and exit\n";
    echo "      --geonames-account=<account>    geonames account to reverse geocode images ( see http://www.geonames.org/export/#ws)\n";
    echo "      --date-time-pattern=<regexp>    only for nerds: RE to extract datetime from filename if no exif data present ( ex. /^(?P<year>[0-9]{4})-(?P<month>[0-9]{2})-(?P<day>[0-9]{2})-(?P<hour>[0-9]{2})(?P<minute>[0-9]{2})(?P<second>[0-9]{2})/ for 2013-07-12-140959.jpg ) \n";
    echo "  -h, --help  Print this help\n";
    die;
  }
  
  $outImage="annotated.jpg";
  $image=false;
  $jpegQuality=90;
  $photo_width=15;
  $photo_height=10;
  $dpi=300;
  $inDir=".";
  $outDir=".";
  $rename=false;
  $scan=false;
  $recursive=false;
  $rasterizer="magick";
  $geonamesAccount=false;

  $templates=array();
  $defaultTemplate="XY";
  $dateTimePattern=false;

  $filez = scandir ("templates");
  foreach ($filez as $file){
    if (preg_match ("/(.*?)\.svg\.php/",$file,$arr)){
      $template = $arr[1];
      if (!$defaultTemplate){
        $defaultTemplate = $template;
      }
      $templates[]=$template;
    }
  }
  $template = $defaultTemplate;

  $longopts  = array(
      "template:",     
      "dpi:",     
      "infile:",  
      "indir:",
      "list-templates",
      "outfile:",
      "outdir:",
      "quality:",    
      "rasterizer:",
      "rename",
      "scan",
      "recursive",
      "recurse",
      "geonames-account:",
      "size:",
      "date-time-pattern:",
      "help"         
  );
  $options = getopt("hi:o:d:q:I:O:t:r:S:", $longopts);
  foreach ($options as $key=>$value){
    switch ($key){
      case 'h':
      case 'help':
        usage();
      case 'list-templates':
        echo "Available templates:\n";
        foreach ($templates as $template){
          echo "\t${template}\n";
        }
        die;

        break;
      case 'geonames-account':
        $geonamesAccount=$value;
        break;
      case 'rename':
        $rename=true;
        break;
      case 'scan':
        $scan=true;
        break;
      case 'recursive':
      case 'recurse':
        $recursive=true;
        break;
      case 'r':
      case 'rasterizer':
        switch ($value){
          case 'batik':
          case 'magick':
            break;
          default:
            usage();
        }
        $rasterizer = $value;        
        break;
      case 'q':
      case 'quality':
        $jpegQuality=(int) $value;
        break;
      case 'S':
      case 'size':
        if ( preg_match ("/(\d+)[xX](\d+)/",$value,$arr) ){
          $photo_width=(int)$arr[2];
          $photo_height=(int)$arr[1];
        } else {
          usage();
        }
        break;

      case 'i':
      case 'infile':
        $image = $value;
        break;
      case 'I':
      case 'indir':
        $inDir = $value;
        $scan=true;
        break;        
      case 'o':
      case 'outfile':
        $outImage = $value;
        break;
      case 'O':
      case 'outdir':
        $outDir = $value;
        break;        

      case 'd':
      case 'dpi':
        $dpi = $value;
        break;
      case 't':
      case 'template':
        $template=$value;
        break;
      case 'date-time-pattern':
        $dateTimePattern=$value;
        break;

    }
  }
  alog ("Annotating using template [$template] with photo size [${photo_height}x${photo_width} cm (height X width )]");
  if ($geonamesAccount) {
    alog ("geonames account [$geonamesAccount]");
  }

  if ($scan){
    doDir ();  	
  	die;
  }  
  if ($image){
  	doImage();
  	die;
	}
	usage();

?>
